@@ -1,11 +1,11 @@
-# $Header: /home/auspex/h/bair/phelps/spine/sb/RCS/searchbox.tcl,v 1.20 1996/06/05 02:09:06 phelps Exp phelps $
+# $Header: /home/orodruin/h/bair/phelps/spine/sb/RCS/searchbox.tcl,v 1.22 1997/08/11 19:01:20 phelps Exp phelps $
 
 #
 # SearchBox mega widget
 #   incremental and regular expression searching in a text widget
 #   for examples of use, see TkMan and tkinfo
 #
-#   Copyright (C) 1993-1996 Tom Phelps (phelps@ACM.org)
+#   Copyright (C) 1993-1997 Tom Phelps (phelps@ACM.org)
 #
 # extracted from TkMan, and then used by TkMan and NBT 6-Aug-93
 #
@@ -28,35 +28,6 @@
 # does NOT require taputils.tcl
 
 
-set sbx(boilerplate) {
-Searchbox
-Copyright (c) 1993-1996  T.A. Phelps
-
-Permission to use, copy, modify, and distribute this software and its
-documentation for educational, research and non-profit purposes, 
-without fee, and without a written agreement is hereby granted, 
-provided that the above copyright notice and the following three 
-paragraphs appear in all copies.  
-
-Permission to incorporate this software into commercial products may 
-be obtained from the Office of Technology Licensing, 2150 Shattuck 
-Avenue, Suite 510, Berkeley, CA  94704. 
-
-IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY 
-FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES 
-ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF 
-THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF 
-SUCH DAMAGE.
-
-THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, 
-INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
-MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE 
-PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF 
-CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, 
-ENHANCEMENTS, OR MODIFICATIONS.
-}
-
-
 ### default key bindings
 # ""=unmodified, A=alt, C=control, M=meta
 # (Shift should not be used as a modifier.  Different keyboards
@@ -66,7 +37,7 @@ ENHANCEMENTS, OR MODIFICATIONS.
 #  capital A, just write "sb(key,-A)" or "sb(key,M-A)" or sb(key,C-A)"--
 #  ignoring the shift altogether.)
 
-set sb(key,*) "add modifiers in this order: M, C, A"
+set sb(key,*) "state multiple modifiers in this order: S, C, M, A"
 
 # Emacs-like
 
@@ -105,16 +76,17 @@ set sb(key,-F27) pagestart
 set sb(key,-F29) pageup
 set sb(key,-F35) pagedown
 set sb(key,-R13) pageend
+set sb(key,-F33) pageend
 
 
 set sbx(wrap-msg) "No more matches; try again to wrap around"
 
-# cursor keys -- built into Tk 4.0
+# cursor keys -- some built into Tk 4.0
 
-#set sb(key,-Prior) pageup
-#set sb(key,-Next) pagedown
-#set sb(key,-Home) pagestart
-#set sb(key,-End) pageend
+set sb(key,-Prior) pageup
+set sb(key,-Next) pagedown
+set sb(key,-Home) pagestart
+set sb(key,-End) pageend
 set sb(key,M-less) pagestart
 set sb(key,M-greater) pageend
 
@@ -150,9 +122,11 @@ if {[info tclversion]<7.4 || $tk_version<4.0} {
 #
 #--------------------------------------------------
 
-proc searchboxSearch {pat regexp casesen tag t {wmsg ""} {wcnt ""}} {
+proc searchboxSearch {pat regexp casesen tag t {wmsg ""} {wcnt ""} args} {
 	global sbx
 
+	if {$args=="{}"} {set args ""}
+
 	if {$casesen==-1} {set casesen [expr {[string tolower $pat]!=$pat}]}
 	set caseopt [expr !$casesen?"-nocase":""]
 	if {$regexp} {set type "regexp"} {set type "exact"}
@@ -162,8 +136,9 @@ proc searchboxSearch {pat regexp casesen tag t {wmsg ""} {wcnt ""}} {
 		return 0
 	}
 
+
 	# check for bad regular expression
-	if {$type=="regexp"&&[catch {regexp $pat bozomaniac}]} {
+	if {![string match "TAG:*" $pat] && $type=="regexp" && [catch {regexp $pat bozomaniac}]} {
 		winstderr $wmsg "Malformed regular expression."
 		return 0
 	}
@@ -172,21 +147,34 @@ proc searchboxSearch {pat regexp casesen tag t {wmsg ""} {wcnt ""}} {
 
 	# wow, heaps faster with Tk 4.0's built-in searching
 	set cnt 0
-	set index 1.0
-	set len 0
-	while {[set index [eval "$t search -forwards -$type $caseopt -count len -- {$pat} {$index+$len chars} end"]]!=""} {
-		$t tag add $tag $index "$index + $len chars"
-		incr cnt
+
+
+	if {[regexp "^TAG:(.*)" $pat all ptag]} {
+		$t tag remove $tag 1.0 end
+		set cnt 0
+		foreach wtag [$t tag names] {
+			if {[string match $ptag $wtag]} {
+				foreach {s e} [$t tag ranges $wtag] {$t tag add $tag $s $e; incr cnt}
+			}
+		}
+
+	} else {
+		set index 1.0
+		set len 0
+		while {[set index [eval "$t search -forwards -$type $caseopt $args -count len -- {$pat} {$index+$len chars} end"]]!=""} {
+			$t tag add $tag $index "$index + $len chars"
+			incr cnt
+		}
 	}
 
 
-	if {$cnt==0} {set txt "no matches"} elseif {$cnt==1} {set txt "$cnt match"} {set txt "$cnt matches"}
+	if {$cnt==0} {set txt "no matches"} else {set txt "$cnt [plural $cnt match es]"}
 #	winstdout $wcnt $txt
-	$wcnt configure -text $txt
+	if {$wcnt!=""} {$wcnt configure -text $txt}
 
 	# show the first match
 	set sbx(ratend$t) 0; set sbx(ratstart$t) 0
-	searchboxNext $tag $t $wmsg 0
+#	searchboxNext $tag $t $wmsg 0
 
 	return $cnt
 }
@@ -210,19 +198,25 @@ proc searchboxNext {tag t {wmsg ""} {yoff ""}} {
 
 	if {[$t tag ranges $tag]==""} { winstderr $wmsg "No matches!"; return }
 
-	if {$yoff==""} {set yoff [winfo height $t]}
-	set hit [$t tag nextrange $tag [$t index @0,$yoff]]
+	if {$yoff==""} {
+		set yoff [winfo height $t]; set start [$t index @0,$yoff]
+#puts "prevrange = [$t tag prevrange search @0,$yoff @0,0]"
+		if {[set prev [$t tag prevrange search @0,$yoff @0,0]]!=""} {set start [$t index [lsecond $prev]+1c]}
+	} else { set start [$t index @0,$yoff] }
+	set hit [$t tag nextrange $tag $start]
 	if {$hit==""} {
 		if {$sbx(ratend$t)} {
 			set hit [$t tag nextrange $tag 1.0]
 			set sbx(ratend$t) 0
 		} else {
-			winstderr $wmsg $sbx(wrap-msg)
 			set sbx(ratend$t) 1
+			if {[$t bbox 1.0]==""} {winstderr $wmsg $sbx(wrap-msg)} {searchboxNext $tag $t $wmsg $yoff}
 		}
 	}
 	if {$hit!=""} {
-		$t see [lindex $hit 0]
+#		$t see [lindex $hit 0]
+		$t mark set hit [lindex $hit 0]
+		$t see hit
 #		winstdout $wmsg ""; # clear wraparound messages
 	}
 }
@@ -238,13 +232,15 @@ proc searchboxPrev {tag t {wmsg ""}} {
 			set hit [$t tag prevrange $tag end]
 			set sbx(ratstart$t) 0
 		} else {
-			winstderr $wmsg $sbx(wrap-msg)
 			set sbx(ratstart$t) 1
+			if {[$t bbox end-1c]==""} {winstderr $wmsg $sbx(wrap-msg)} {searchboxPrev $tag $t $wmsg}
 		}
 	}
 
 	if {$hit!=""} {
-		$t see [lindex $hit 0]
+#		$t see [lindex $hit 0]
+		$t mark set hit [lindex $hit 0]
+		$t see hit
 #		winstdout $wmsg ""; # clear wraparound messages
 	}
 }
@@ -266,11 +262,13 @@ proc searchboxPrev {tag t {wmsg ""}} {
 #
 #--------------------------------------------------
 
-proc searchboxKeyNav {m k casesen t {wmsg ""} {firstmode 0}} {
+proc searchboxKeyNav {m k casesen t {wmsg ""} {firstmode 0} args} {
 #puts "$m $k $casesen $t $wmsg $firstmode"
 	global sb sbx
 
-	if {[regexp {(Shift|Control|Meta|Alt)_.} $k]} {return 0}
+	if {$args=="{}"} {set args ""}
+
+	if {[regexp {(Shift|Control|Meta|Alt|NunLock)_.} $k]} {return 0}
 	if {![info exists sbx(try$t)]} {
 		set sbx(try$t) 0
 		set sbx(vect$t) 1
@@ -377,7 +375,7 @@ proc searchboxKeyNav {m k casesen t {wmsg ""} {firstmode 0}} {
 			set dir [expr {$sbx(vect$t)==1?"-forwards":"-backwards"}]
 			set type [expr $firstmode?"-regexp":"-exact"]
 			set pfx [expr $firstmode?"^":""]
-			set search "$t search $dir $type $caseopt -count klen -- {$pfx$keys} $start"
+			set search "$t search $dir $type $caseopt -count klen $args -- {$pfx$keys} $start"
 #DEBUG {puts "$search ... $end"}
 			set found [eval "$search $end"]
 			set anywhere [eval $search]
@@ -389,7 +387,8 @@ proc searchboxKeyNav {m k casesen t {wmsg ""} {firstmode 0}} {
 				if {$found!=""} {
 					$t tag remove isearch 1.0 end
 					$t tag add isearch $found "$found + $klen c"
-					$t see $found
+					$t mark set hit $found
+					$t see hit
 #					winstdout $wmsg ""; # clear wraparound messages(?)
 					set sbx(iatstart$t) 0; set sbx(iatend$t) 0
 				} else {
